"""
Collaborative text editing plugin for gedit.

Copyright: Matt Colyer, 2007
License: GPLv2
"""

import copy
import gtk
import gobject
import re
import time
import xml.dom.minidom
import xml.dom.ext
import md5
import dbus
import telepathy

# Connect dbus to the glib mainloop
from dbus.mainloop.glib import DBusGMainLoop
DBusGMainLoop(set_as_default=True)

MISSION_CONTROL_BUS = 'org.freedesktop.Telepathy.Houston.MissionControl'
MISSION_CONTROL_BUS_PATH = '/org/freedesktop/Telepathy/Houston/MissionControl'

try:
    import gedit
    
    class CollaboratePlugin(gedit.Plugin):
        """The base class which configures the Collaborate plugin. It handles
        creating a seperate CollaborateWindowHelper for each available gedit
        window. It is also responsible for deactiving and removing those
        objects as gedit windows are closed. """

        def __init__(self):
            gedit.Plugin.__init__(self)
            self._instances = {}

        def activate(self, window):
            """When a gedit window is created, create a new CollaborateWindowHelper."""
            self._instances[window] = CollaborateWindowHelper(self, window)
        
        def deactivate(self, window):
            """When a gedit window is closed, destroy its respective CollaborateWindowHelper."""
            self._instances[window].deactivate()
            del self._instances[window]

        def update_ui(self, window):
            """Pass update_ui calls through to the appropriate CollaborateWindowHelper."""
            self._instances[window].update_ui()

        def is_configurable(self):
            """Inform gedit that the Collaborate plugin is currently not configurable."""
            return false
        
    class CollaborateWindowHelper:
        """Represents Collaborate's functionality which belongs on the window
        level. Currently this includes creating the menuitems for each
        GeditWindow."""

        def __init__(self, plugin, window):
            self._window = window
            self._plugin = plugin
            self._documents = {}

            self._insert_menu()

            self._window.connect('tab-removed', self.on_tab_removed)
            
        def deactivate(self):
            """Cleanup class resources and prepare for destruction."""
            self._remove_menu()

            self._window = None
            self._plugin = None
            self._action_group = None

            for document in documents:
                document.deactivate()

            del documents

        def _insert_menu(self):
            """Add menu items and actions to the current gedit window."""
            manager = self._window.get_ui_manager()

            # Create a new action group
            self._action_group = gtk.ActionGroup('CollaboratePluginActions')
            self._action_group.add_toggle_actions([('Share', None, 'Share document',
                                             None, 'Share the current document',
                                             self.on_share_document_activate)])
            self._action_group.add_actions([('Join', None, 'Join a document',
                                             None, 'Join a shared document',
                                             self.on_join_document_activate)])

            # Connect actions
            manager.insert_action_group(self._action_group, -1)
            
            # Create and add the Collaborate menuitems
            ui_str = """<ui>
                  <menubar name="MenuBar">
                    <menu name="ToolsMenu" action="Tools">
                      <placeholder name="ToolsOps_2">
                        <menuitem name="CollaborateShare" action="Share"/>
                        <menuitem name="CollaborateJoin" action="Join"/>
                        <separator/>
                      </placeholder>
                    </menu>
                  </menubar>
                </ui>
                """
            self._ui_id = manager.add_ui_from_string(ui_str)

        def _remove_menu(self):
            """Remove Collaborate menu items and actions from the gedit
            window."""
            manager = self._window.get_ui_manager()
            manager.remove_ui(self._ui_id)
            manager.remove_action_group(self._action_group)
            manager.ensure_update()

        def update_ui(self):
            """Updates the user interface components. Currently it sets the
            action group sensitive when the window has an open document."""
            self._action_group.set_sensitive(self._window.get_active_document() != None)

        def on_tab_removed(self, window, tab, data=None):
            """Callback which is activated when a tab is removed. Destroys
            related CollaborateDocumentHelper if it exists."""
            gedit_document = tab.get_document()
            if self._documents.has_key(gedit_document):
                self._documents[gedit_document].deactivate()
                del self._documents[gedit_document]

        def on_share_document_activate(self, action):
            """Callback which is activated by clicking the share document
            menuitem. Shares the current document."""
            if action.get_active():
                action.set_active(True)
                self.create_document_helper()
            else:
                action.set_active(False)
                gedit_document = self._window.get_active_document()
                if self._documents.has_key(gedit_document):
                    self._documents[gedit_document].deactivate()
                    del self._documents[gedit_document]


        def on_join_document_activate(self, action):
            """Callback which is activated by clicking on the join a document
            menuitem. Creates a dialog to ask for the document address."""
            CollaborateDocumentAddressDialog(parent=self._window, window_helper=self) 

        def create_document_helper(self, document_address=None):
            """Helper function which creates a CollaborateDocumentHelper for
            the active document."""
            gedit_document = self._window.get_active_document()
            gedit_view = self._window.get_active_view()

            # Checks to make sure there is a gedit document available
            if not gedit_document:
                return
            
            document = CollaborateDocumentHelper(self._plugin, self._window, gedit_document, gedit_view, document_address)
            # FIXME: Check to make sure that the document could be shared *This doesn't work correctly yet*
            if document:
                self._documents[gedit_document] = document
 
    class CollaborateDocumentAddressDialog(gtk.Dialog):
        def __init__(self, parent=None, window_helper=None):
            self._window = parent
            self._window_helper = window_helper

            gtk.Dialog.__init__(self, title='Document Address', 
                                parent=self._window,
                                buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT, gtk.STOCK_OK, gtk.RESPONSE_ACCEPT),
                                flags=gtk.DIALOG_MODAL,
                                )
            entry = gtk.Entry()
            entry.set_activates_default(True)
            self.set_default_response(gtk.RESPONSE_ACCEPT)
            self.vbox.pack_start(entry, padding=6)
            self.connect('response', self.response, entry)
            self.show_all()

        def response(self, dialog, response_id, data=None):
            #FIXME: Validate address input
            document_address = data.get_text()
            dialog.destroy()
            
            if response_id == gtk.RESPONSE_ACCEPT:
                self._window.create_tab(True)
                self._window_helper.create_document_helper(document_address)

    class CollaborateDocumentHelper:
        """Represents Collaborate's functionality which belongs on the document
        level. Currently this means intercepting document changes and
        broadcasting them to fellow collaborators as well as listening for
        changes from other collaborators."""

        def __init__(self, plugin, window, doc, view, remote_channel_address=None):
            self._plugin = plugin
            self._window = window
            self._gedit_document = doc
            self._gedit_view = view

            # Used to store participant's text tags
            self._participants = {}

            # See if we are joining an existing channel or if we need to create our own.
            if remote_channel_address is None:
                remote_channel_address = "%s@conference.jabber.org" % self.hash()
                sharing = True
            else:
                sharing = False

            # Ask Telepathy if there are any jabber accounts available.
            bus = dbus.SessionBus()
            mission_control = bus.get_object(MISSION_CONTROL_BUS, MISSION_CONTROL_BUS_PATH)
            connected_accounts  = mission_control.GetConnectedAccounts()
            jabber_connected_accounts = filter(lambda x: x[3] == 'Jabber', connected_accounts)
            
            # Act according to how many accounts are available.
            if len(jabber_connected_accounts) == 0:
                # No accounts are available
                CollaborateAccountMessageDialog(parent=self._window)

                #FIXME: This won't work
                return None

            elif len(jabber_connected_accounts) == 1:
                # Use the only existing account

                # Get the telepathy connection
                proxy_object = bus.get_object(jabber_connected_accounts[0][0], jabber_connected_accounts[0][1])
                connection = dbus.Interface(proxy_object, telepathy.interfaces.CONN_INTERFACE)

                # Connect to the jabber chat room
                handle = connection.RequestHandles(telepathy.CONNECTION_HANDLE_TYPE_ROOM, [remote_channel_address])[0]
                transport_path = connection.RequestChannel(telepathy.interfaces.CHANNEL_TYPE_TEXT, telepathy.CONNECTION_HANDLE_TYPE_ROOM, handle, True)
                proxy_object = bus.get_object(jabber_connected_accounts[0][0], transport_path)
                self._transport = dbus.Interface(proxy_object, telepathy.interfaces.CHANNEL_TYPE_TEXT)
            else:
                # TODO: If >1 then allow them to choose one
                pass

            # Show a dialog with the document's name if the user is sharing a document
            if sharing:
                CollaborateShareMessageDialog(parent=self._window, remote_channel_address=remote_channel_address)
               
            # Connect up the received messages to be parsed.
            self._transport.connect_to_signal("Received", self.handle_message)

            # Connect up document signals
            self._insert_signal_id = self._gedit_document.connect('insert-text', self.on_insert)
            self._delete_signal_id = self._gedit_document.connect('delete-range', self.on_delete)

        def deactivate(self):
            """Cleanup class resources and prepare for destruction."""
            self._plugin = None
            self._window = None
            self._gedit_document = None
            self._gedit_view = None
            self._participants = None
                
            self._transport = dbus.Interface(self._transport, telepathy.interfaces.CHANNEL_INTERFACE)
            self._transport.Close()
            self._transport = None
            pass
        
        def handle_message(self, id, timestamp, sender, type, flags, text):
            """Callback which handles incoming network messages."""
            try:
                # Attempt to parse the message
                operation = Operation.parse(str(text))

                # Lock the document from editing and remove callbacks
                self._gedit_view.set_property('editable', False)
                self._gedit_document.disconnect(self._insert_signal_id)
                self._gedit_document.disconnect(self._delete_signal_id)

                # Perform the operation
                self.execute(operation, sender)

                # Add the callbacks and allow editing again
                self._insert_signal_id = self._gedit_document.connect('insert-text', self.insert)
                self._delete_signal_id = self._gedit_document.connect('delete-range', self.delete)
                self._gedit_view.set_property('editable', True)

            except InvalidOperation:
                # Ignore invalid messages for now
                pass

        def execute(self, command, participant):
            """Applies a command to the current document."""
            # FIXME: Should really use the tag table instead of my own hash table
            if not self._participants.has_key(participant):
                self._participants[participant] = self._gedit_document.create_tag()
                
                #FIXME: Get more creative here, limited to only 6 colors
                self._participants[participant].set_property('background', 
                        ['#FFF595', '#ADC7E3', '#C9F3A0', '#FFD394', '#D4BCD2','#F89999'][len(self._participants) % 7])

            if isinstance(command, InsertOperation):
                self._gedit_document.insert_with_tags(self._gedit_document.get_iter_at_offset(command.p), command.s, self._participants[participant])
            elif isinstance(command, DeleteOperation):
                self._gedit_document.delete(self._gedit_document.get_iter_at_offset(command.p), self._gedit_document.get_iter_at_offset(command.p+command.l))

        def on_insert(self, textbuffer, iter, text, length, data=None):
            """Callback which intercepts local document inserts."""
            self._transport.Send(telepathy.CHANNEL_TEXT_MESSAGE_TYPE_NORMAL, str(InsertOperation(text, iter.get_offset()))+"\n")
        
        def on_delete(self, textview, start, end, data=None):
            """Callback which intercepts local document deletes."""
            offset = start.get_offset()
            length = end.get_offset() - offset
            self._transport.Send(telepathy.CHANNEL_TEXT_MESSAGE_TYPE_NORMAL, str(DeleteOperation(length, offset)))

        def hash(self):
            """Creates an MD5 based on the document's local filename. If it doesn't have a filename
            use the current time"""
            if self._gedit_document.get_uri():
                return md5.new(self._gedit_document.get_uri()).hexdigest()
            else:
                return md5.new(str(time.time())).hexdigest()
               
    class CollaborateAccountMessageDialog(gtk.MessageDialog):
        def __init__(self, parent=None):
            self._window = parent

            gtk.MessageDialog.__init__(self, 
                                parent=self._window,
                                buttons=gtk.BUTTONS_CLOSE,
                                flags=gtk.DIALOG_MODAL,
                                message_format="No Telepathy Jabber Account"
                                )
            self.format_secondary_markup("You must have a connected Jabber account in Telepathy before you are able to share a document. A dialog will appear following this message which will allow you to modify your Telepathy accounts.")
            self.connect('response', self.response)
            self.show()
 
        def response(self, dialog, response_id, data=None):
            """Callback for the account message dialog. It destroys the dialog
            and shows the Telepathy accounts dialog."""
            dialog.destroy()
            bus = dbus.SessionBus()
            mission_control = bus.get_object(MISSION_CONTROL_BUS, MISSION_CONTROL_BUS_PATH)
            mission_control.ShowAccounts()

    class CollaborateShareMessageDialog(gtk.MessageDialog):
        def __init__(self, parent=None, remote_channel_address=""):
            self._window = parent

            gtk.MessageDialog.__init__(self, 
                                parent=self._window,
                                buttons=gtk.BUTTONS_CLOSE,
                                flags=gtk.DIALOG_MODAL,
                                message_format="Document is Shared"
                                )
            self.format_secondary_markup("Your document is shared at <b>%s</b> Copy and paste this address to other participants." % remote_channel_address)
            self.connect('response', self.response)
            self.show()
 
        def response(self, dialog, response_id, data=None):
            dialog.destroy()
 
except:
    print "WARNING: No Gedit environment"

class InvalidOperation(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class Operation:
    @staticmethod
    def parse(string):
        insert_re = re.compile("Insert\[\"([^\"]+)\",([0-9]+)\]")
        delete_re = re.compile("Delete\[\"([^\"]+)\",([0-9]+)\]")
        if insert_re.match(string):
            text, position = insert_re.match(string).groups()
            position = int(position)
            return InsertOperation(text, position)
        elif delete_re.match(string):
            length, position = delete_re.match(string).groups()
            length = int(length)
            position = int(position)
            return DeleteOperation(length, position)
        else:
            raise InvalidOperation, "Invalid message string"

    def is_insert(self):
        return False
    
    def is_delete(self):
        return false
    
    def include(op_a, op_b):
        if op_a.is_insert() and op_b.is_insert(): op_ap = op_a.include_ii(op_b)
        elif op_a.is_insert() and op_b.is_delete(): op_ap = op_a.include_id(op_b)
        elif op_a.is_delete() and op_b.is_insert(): op_ap = op_a.include_di(op_b)
        elif op_a.is_delete() and op_b.is_delete(): op_ap = op_a.include_dd(op_b)
        return op_ap
    
    def include_ii(op_a, op_b):
        assert(op_a.is_insert())
        assert(op_b.is_insert())
        
        if (op_a.p < op_b.p): return copy.deepcopy(op_a)
        else: return InsertOperation(op_a.s, op_a.p + op_b.l)
    
    def include_id(op_a, op_b):
        assert(op_a.is_insert())
        assert(op_b.is_delete())
        
        if (op_a.p <= op_b.p): return copy.deepcopy(op_a)
        elif (op_a.p > (op_b.p + op_b.l)): return InsertOperation(op_a.s,
            op_a.p - op_b.l)
        else:
            op_ap = InsertOperation(op_a.s, op_b.p)
            op_ap.save_li(op_a, op_b)
            return op_ap
    
    def include_di(op_a, op_b):
        if (op_b.p >= (op_a.p + op_a.l)): return copy.deepcopy(op_a)
        elif (op_a.p >= op_b.p): return DeleteOperation(op_a.l, op_a.p +
                op_b.l)
        else:
            op_return = DeleteOperation(op_b.p - op_a.p, op_a.p)
            op_return.set_sub(op_a.l - (op_b.p - op_a.p), op_b.p + op_b.l)
            return op_return
    
    def include_dd(op_a, op_b):
        if (op_b.p >= (op_a.p + op_a.l)): return copy.deepcopy(op_a)
        elif (op_a.p >= (op_b.p + op_b.l)): return DeleteOperation(op_a.l,
                op_a.p - op_b.l)
        else:
            if ((op_b.p <= op_a.p) and ((op_a.p + op_a.l) <= (op_b.p + op_b.l))): 
                op_return = DeleteOperation(0, op_a.p)
            elif ((op_b.p <= op_a.p) and ((op_a.p + op_a.l) > (op_b.p + op_b.l))): 
                op_return = DeleteOperation(op_a.p + op_a.l - (op_b.p + op_b.l), op_b.p)
            elif ((op_b.p > op_a.p) and ((op_b.p + op_b.l) >= (op_a.p + op_a.l))):
                op_return = DeleteOperation(op_b.p - op_a.p, op_a.p)
            else:
                op_return = DeleteOperation(op_a.l - op_b.l, op_a.p)
            op_return.save_li(op_a, op_b)
            return op_return
    
    def exclude(op_a, op_b):
        if op_a.is_insert() and op_b.is_insert(): op_ap = op_a.exclude_ii(op_b)
        elif op_a.is_insert() and op_b.is_delete(): op_ap = op_a.exclude_id(op_b)
        elif op_a.is_delete() and op_b.is_insert(): op_ap = op_a.exclude_di(op_b)
        elif op_a.is_delete() and op_b.is_delete(): op_ap = op_a.exclude_dd(op_b)
        return op_ap
    
    def exclude_ii(op_a, op_b):
        if (op_a.p < op_b.p): return copy.deepcopy(op_a)
        elif (op_a.p > (op_b.p + op_b.l)): return InsertOperation(op_a.s, op_a.p - op_b.l)
        else:
            op_return = InsertOperation(op_a.s, op_a.p - op_b.p)
            op_return.save_ra(op_b)
            return op_return

    def exclude_id(op_a, op_b):
        if (op_a.check_li(op_b)): return op_a.recover()
        elif (op_a.p < op_b.p): return copy.deepcopy(op_a)
        else: return InsertOperation(op_a.s, op_a.p + op_b.l)

    def exclude_di(op_a, op_b):
        if ((op_a.p + op_a.l) < op_b.p): return copy.deepcopy(op_a)
        elif (op_a.p >= (op_b.p + op_b.l)): return DeleteOperation(op_a.l, op_a.p - op_b.l)
        else:  
            if ((op_b.p <= op_a.p) and ((op_a.p + op_a.l) <= (op_b.p + op_b.l))):
                op_return = DeleteOperation(op_a.l, op_a.p - op_b.p)
            elif ((op_b.p <= op_a.p) and ((op_a.p + op_a.l) > (op_b.p + op_b.l))):
                op_return = DeleteOperation(op_b.p+op_b.l - op_a.p, op_a.p - op_b.p)
                op_return.set_sub((op_a.p+op_a.l) - (op_b.p+op_b.l), op_b.p)
            elif ((op_a.p < op_b.p) and ((op_b.p + op_b.l) <= (op_a.p + op_a.l))):
                op_return = DeleteOperation(op_b.l, 0)
                op_return.set_sub(op_a.l - op_b.l, op_a.p)
            else:
                op_return = DeleteOperation(op_a.l+op_a.p - op_b.p, 0)
                op_return.set_sub(op_b.p - op_a.p, op_a.p)
            op_return.save_ra(op_b)
            return op_return
    
    def exclude_dd(op_a, op_b):
        if op_a.check_li(op_b): return op_a.recover_li(op_b)
        elif (op_b.p >= (op_a.p + op_a.l)): return copy.deepcopy(op_a)
        elif (op_a.p >= op_b.p): return DeleteOperation(op_a.l, op_a.p + op_b.l)
        else:
            op_return = DeleteOperation(op_b.p - op_a.p, op_a.p)
            op_return.set_sub(op_a.l - (op_b.p - op_a.p), op_b.p + op_b.l)
            return op_return
        
    def equals(self, op):
        #FIXME: should throw exception
        print "Not implemented"
    
    def save_ra(self, op):
        self._relative = op

    def check_ra(self, op):
        return self._relative == op

    def check_li(self, op):
        return self._backup.has_key(op)

    def __hash__(self):
        #FIXME: do something smart here
        return 0

class DeleteOperation(Operation):
    def __init__(self, length, position):
        assert(type(length) is int)
        assert(type(position) is int)
        self.p = position
        self.l = length
        self._backup = {}
    
    def __str__(self):
        return "Delete[%d,%d]" % (self.p, self.l)
    
    def is_delete(self):
        return True
    
    def __eq__(self, op):
        return self.equals(op)

    def equals(self, op):
        return self.p == op.p and self.l == op.l
    
    def set_sub(self, length, position):
        """Save a sub delete operation."""
        self.sl = length
        self.sp = position
        self._sub = True
    
    def is_split(self):
        return self._sub
    
    def save_li(self, op_a, op_b):
        assert(op_a.is_delete())
        self._backup[op_b] = (op_a.l, op_a.p)
    
    def recover_li(self, op):
        assert(self._backup.has_key(op))
        self.l, self.p = self._backup[op]

class InsertOperation(Operation):
    def __init__(self, text, position):
        assert(type(text) is str)
        assert(type(position) is int)
        self.s = text
        self.p = position
        self.l = len(text)
        self._backup = {}
    
    def __str__(self):
        return "Insert[\"%s\",%d]" % (self.s, self.p)
    
    def is_insert(self):
        return True
    
    def __eq__(self, op):
        return self.equals(op)

    def equals(self, op):
        return self.s == op.s and self.p == op.p and self.l == op.l
    
    def save_li(self, op_a, op_b):
        assert(op_a.is_insert())
        self._backup[op_b] = (op_a.s, op_a.p)

    def recover_li(self, op):
        assert(self._backup.has_key(op))
        self.l, self.p = self._backup[op]
 
